import { get_session, get_user, PRECHECK_TABLE, require_user } from '../../../../utils/prechecks.ts';
import parse_body from '../../../../utils/bodyparser.ts';
import * as CANNED_RESPONSES from '../../../../utils/canned_responses.ts';
import { ROOM, ROOMS } from '../../../../models/room.ts';

export const PRECHECKS: PRECHECK_TABLE = {};

// GET /api/rooms/:id - Get a room
PRECHECKS.GET = [get_session, get_user, require_user, async (_req: Request, meta: Record<string, any>): Promise<Response | undefined> => {
	const room_id: string = meta.params?.room_id?.toLowerCase().trim() ?? '';

	// lurid is 49 chars as we use them, eg: "also-play-flow-want-form-wide-thus-work-burn-same"
	const room: ROOM | null = room_id.length === 49 ? await ROOMS.get(room_id) : null;

	if (!room) {
		return CANNED_RESPONSES.not_found();
	}

	meta.room = room;
	const room_is_public = room.permissions.read.length === 0;
	const user_has_read_for_room = room_is_public || room.permissions.read.includes(meta.user.id);

	if (!user_has_read_for_room) {
		return CANNED_RESPONSES.permission_denied();
	}
}];
export function GET(_req: Request, meta: Record<string, any>): Response {
	return Response.json(meta.room, {
		status: 200
	});
}

// PUT /api/rooms/:id - Update room
PRECHECKS.PUT = [get_session, get_user, require_user, async (_req: Request, meta: Record<string, any>): Promise<Response | undefined> => {
	const room_id: string = meta.params?.room_id?.toLowerCase().trim() ?? '';

	// lurid is 49 chars as we use them, eg: "also-play-flow-want-form-wide-thus-work-burn-same"
	const room: ROOM | null = room_id.length === 49 ? await ROOMS.get(room_id) : null;

	if (!room) {
		return CANNED_RESPONSES.not_found();
	}

	meta.room = room;
	const user_has_write_for_room = room.permissions.write.includes(meta.user.id);

	if (!user_has_write_for_room) {
		return CANNED_RESPONSES.permission_denied();
	}
}];
export async function PUT(req: Request, meta: Record<string, any>): Promise<Response> {
	const now = new Date().toISOString();

	try {
		const body = await parse_body(req);
		const updated = {
			...meta.room,
			...body,
			id: meta.room.id,
			timestamps: {
				created: meta.room.timestamps.created,
				updated: now
			}
		};

		await ROOMS.update(updated);
		return Response.json(updated, {
			status: 200
		});
	} catch (err) {
		return Response.json({
			error: {
				message: (err as Error)?.message ?? 'Unknown error due to invalid data.',
				cause: (err as Error)?.cause ?? 'invalid_data'
			}
		}, {
			status: 400
		});
	}
}

// DELETE /api/rooms/:id - Delete room
PRECHECKS.DELETE = [
	get_session,
	get_user,
	require_user,
	async (_req: Request, meta: Record<string, any>): Promise<Response | undefined> => {
		const room_id: string = meta.params?.room_id?.toLowerCase().trim() ?? '';

		// lurid is 49 chars as we use them, eg: "also-play-flow-want-form-wide-thus-work-burn-same"
		const room: ROOM | null = room_id.length === 49 ? await ROOMS.get(room_id) : null;

		if (!room) {
			return CANNED_RESPONSES.not_found();
		}

		meta.room = room;
		const user_has_write_for_room = room.permissions.write.includes(meta.user.id);

		if (!user_has_write_for_room) {
			return CANNED_RESPONSES.permission_denied();
		}
	}
];
export async function DELETE(_req: Request, meta: Record<string, any>): Promise<Response> {
	await ROOMS.delete(meta.room);

	return Response.json({
		deleted: true
	}, {
		status: 200
	});
}
